@@ -53,10 +53,6 @@
  * diff -u /tmp/std_find /tmp/bb_find && echo Identical
  */
 
-//applet:IF_FIND(APPLET_NOEXEC(find, find, _BB_DIR_USR_BIN, _BB_SUID_DROP, find))
-
-//kbuild:lib-$(CONFIG_FIND) += find.o
-
 //config:config FIND
 //config:	bool "find"
 //config:	default y
@@ -112,11 +108,11 @@
 //config:	  This option allows find to restrict searches to a single filesystem.
 //config:
 //config:config FEATURE_FIND_MAXDEPTH
-//config:	bool "Enable -maxdepth N"
+//config:	bool "Enable -mindepth N and -maxdepth N"
 //config:	default y
 //config:	depends on FIND
 //config:	help
-//config:	  This option enables -maxdepth N option.
+//config:	  This option enables -mindepth N and -maxdepth N option.
 //config:
 //config:config FEATURE_FIND_NEWER
 //config:	bool "Enable -newer: compare file modification times"
@@ -124,7 +120,7 @@
 //config:	depends on FIND
 //config:	help
 //config:	  Support the 'find -newer' option for finding any files which have
-//config:	  a modified time that is more recent than the specified FILE.
+//config:	  modification time that is more recent than the specified FILE.
 //config:
 //config:config FEATURE_FIND_INUM
 //config:	bool "Enable -inum: inode number matching"
@@ -230,11 +226,118 @@
 //config:	help
 //config:	  Support the 'find -links' option for matching number of links.
 
+//applet:IF_FIND(APPLET_NOEXEC(find, find, BB_DIR_USR_BIN, BB_SUID_DROP, find))
+
+//kbuild:lib-$(CONFIG_FIND) += find.o
+
+//usage:#define find_trivial_usage
+//usage:       "[PATH]... [OPTIONS] [ACTIONS]"
+//usage:#define find_full_usage "\n\n"
+//usage:       "Search for files and perform actions on them.\n"
+//usage:       "First failed action stops processing of current file.\n"
+//usage:       "Defaults: PATH is current directory, action is '-print'\n"
+//usage:     "\n	-follow		Follow symlinks"
+//usage:	IF_FEATURE_FIND_XDEV(
+//usage:     "\n	-xdev		Don't descend directories on other filesystems"
+//usage:	)
+//usage:	IF_FEATURE_FIND_MAXDEPTH(
+//usage:     "\n	-maxdepth N	Descend at most N levels. -maxdepth 0 applies"
+//usage:     "\n			actions to command line arguments only"
+//usage:     "\n	-mindepth N	Don't act on first N levels"
+//usage:	)
+//usage:	IF_FEATURE_FIND_DEPTH(
+//usage:     "\n	-depth		Act on directory *after* traversing it"
+//usage:	)
+//usage:     "\n"
+//usage:     "\nActions:"
+//usage:	IF_FEATURE_FIND_PAREN(
+//usage:     "\n	( ACTIONS )	Group actions for -o / -a"
+//usage:	)
+//usage:	IF_FEATURE_FIND_NOT(
+//usage:     "\n	! ACT		Invert ACT's success/failure"
+//usage:	)
+//usage:     "\n	ACT1 [-a] ACT2	If ACT1 fails, stop, else do ACT2"
+//usage:     "\n	ACT1 -o ACT2	If ACT1 succeeds, stop, else do ACT2"
+//usage:     "\n			Note: -a has higher priority than -o"
+//usage:     "\n	-name PATTERN	Match file name (w/o directory name) to PATTERN"
+//usage:     "\n	-iname PATTERN	Case insensitive -name"
+//usage:	IF_FEATURE_FIND_PATH(
+//usage:     "\n	-path PATTERN	Match path to PATTERN"
+//usage:     "\n	-ipath PATTERN	Case insensitive -path"
+//usage:	)
+//usage:	IF_FEATURE_FIND_REGEX(
+//usage:     "\n	-regex PATTERN	Match path to regex PATTERN"
+//usage:	)
+//usage:	IF_FEATURE_FIND_TYPE(
+//usage:     "\n	-type X		File type is X (one of: f,d,l,b,c,...)"
+//usage:	)
+//usage:	IF_FEATURE_FIND_PERM(
+//usage:     "\n	-perm MASK	At least one mask bit (+MASK), all bits (-MASK),"
+//usage:     "\n			or exactly MASK bits are set in file's mode"
+//usage:	)
+//usage:	IF_FEATURE_FIND_MTIME(
+//usage:     "\n	-mtime DAYS	mtime is greater than (+N), less than (-N),"
+//usage:     "\n			or exactly N days in the past"
+//usage:	)
+//usage:	IF_FEATURE_FIND_MMIN(
+//usage:     "\n	-mmin MINS	mtime is greater than (+N), less than (-N),"
+//usage:     "\n			or exactly N minutes in the past"
+//usage:	)
+//usage:	IF_FEATURE_FIND_NEWER(
+//usage:     "\n	-newer FILE	mtime is more recent than FILE's"
+//usage:	)
+//usage:	IF_FEATURE_FIND_INUM(
+//usage:     "\n	-inum N		File has inode number N"
+//usage:	)
+//usage:	IF_FEATURE_FIND_USER(
+//usage:     "\n	-user NAME/ID	File is owned by given user"
+//usage:	)
+//usage:	IF_FEATURE_FIND_GROUP(
+//usage:     "\n	-group NAME/ID	File is owned by given group"
+//usage:	)
+//usage:	IF_FEATURE_FIND_SIZE(
+//usage:     "\n	-size N[bck]	File size is N (c:bytes,k:kbytes,b:512 bytes(def.))"
+//usage:     "\n			+/-N: file size is bigger/smaller than N"
+//usage:	)
+//usage:	IF_FEATURE_FIND_LINKS(
+//usage:     "\n	-links N	Number of links is greater than (+N), less than (-N),"
+//usage:     "\n			or exactly N"
+//usage:	)
+//usage:	IF_FEATURE_FIND_CONTEXT(
+//usage:     "\n	-context CTX	File has specified security context"
+//usage:	)
+//usage:	IF_FEATURE_FIND_PRUNE(
+//usage:     "\n	-prune		If current file is directory, don't descend into it"
+//usage:	)
+//usage:     "\nIf none of the following actions is specified, -print is assumed"
+//usage:     "\n	-print		Print file name"
+//usage:	IF_FEATURE_FIND_PRINT0(
+//usage:     "\n	-print0		Print file name, NUL terminated"
+//usage:	)
+//usage:	IF_FEATURE_FIND_EXEC(
+//usage:     "\n	-exec CMD ARG ;	Run CMD with all instances of {} replaced by"
+//usage:     "\n			file name. Fails if CMD exits with nonzero"
+//usage:	)
+//usage:	IF_FEATURE_FIND_DELETE(
+//usage:     "\n	-delete		Delete current file/directory. Turns on -depth option"
+//usage:	)
+//usage:
+//usage:#define find_example_usage
+//usage:       "$ find / -name passwd\n"
+//usage:       "/etc/passwd\n"
+
 #include <fnmatch.h>
 #include "libbb.h"
 #if ENABLE_FEATURE_FIND_REGEX
-#include "xregex.h"
+# include "xregex.h"
 #endif
+/* GNUism: */
+#ifndef FNM_CASEFOLD
+# define FNM_CASEFOLD 0
+#endif
+
+#define dbg(...) ((void)0)
+/* #define dbg(...) bb_error_msg(__VA_ARGS__) */
 
 /* This is a NOEXEC applet. Be very careful! */
 
@@ -256,7 +359,7 @@ typedef struct {
 
                         ACTS(print)
                         ACTS(name,  const char *pattern; bool iname;)
-IF_FEATURE_FIND_PATH(   ACTS(path,  const char *pattern;))
+IF_FEATURE_FIND_PATH(   ACTS(path,  const char *pattern; bool ipath;))
 IF_FEATURE_FIND_REGEX(  ACTS(regex, regex_t compiled_pattern;))
 IF_FEATURE_FIND_PRINT0( ACTS(print0))
 IF_FEATURE_FIND_TYPE(   ACTS(type,  int type_mask;))
@@ -278,8 +381,12 @@ IF_FEATURE_FIND_LINKS(  ACTS(links, char links_char; int links_count;))
 struct globals {
 	IF_FEATURE_FIND_XDEV(dev_t *xdev_dev;)
 	IF_FEATURE_FIND_XDEV(int xdev_count;)
+#if ENABLE_FEATURE_FIND_MAXDEPTH
+	int minmaxdepth[2];
+#endif
 	action ***actions;
-	bool need_print;
+	smallint need_print;
+	smallint xdev_on;
 	recurse_flags_t recurse_flags;
 } FIX_ALIASING;
 #define G (*(struct globals*)&bb_common_bufsiz1)
@@ -288,7 +395,8 @@ struct globals {
 		char G_sizecheck[sizeof(G) > COMMON_BUFSIZE ? -1 : 1]; \
 	}; \
 	/* we have to zero it out because of NOEXEC */ \
-	memset(&G, 0, offsetof(struct globals, need_print)); \
+	memset(&G, 0, sizeof(G)); \
+	IF_FEATURE_FIND_MAXDEPTH(G.minmaxdepth[1] = INT_MAX;) \
 	G.need_print = 1; \
 	G.recurse_flags = ACTION_RECURSE; \
 } while (0)
@@ -364,14 +472,32 @@ static int exec_actions(action ***appp, const char *fileName, const struct stat
 #if ENABLE_FEATURE_FIND_NOT
 			if (ap->invert) rc ^= TRUE;
 #endif
+			dbg("grp %d action %d rc:0x%x", cur_group, cur_action, rc);
 			if (rc & TRUE) /* current group failed, try next */
 				break;
 		}
 	}
+	dbg("returning:0x%x", rc ^ TRUE);
 	return rc ^ TRUE; /* restore TRUE bit */
 }
 
 
+#if !FNM_CASEFOLD
+static char *strcpy_upcase(char *dst, const char *src)
+{
+	char *d = dst;
+	while (1) {
+		unsigned char ch = *src++;
+		if (ch >= 'a' && ch <= 'z')
+			ch -= ('a' - 'A');
+		*d++ = ch;
+		if (ch == '\0')
+			break;
+	}
+	return dst;
+}
+#endif
+
 ACTF(name)
 {
 	const char *tmp = bb_basename(fileName);
@@ -387,13 +513,25 @@ ACTF(name)
 	 * but somewhere between 4.1.20 and 4.4.0 GNU find stopped using it.
 	 * find -name '*foo' should match .foo too:
 	 */
+#if FNM_CASEFOLD
 	return fnmatch(ap->pattern, tmp, (ap->iname ? FNM_CASEFOLD : 0)) == 0;
+#else
+	if (ap->iname)
+		tmp = strcpy_upcase(alloca(strlen(tmp) + 1), tmp);
+	return fnmatch(ap->pattern, tmp, 0) == 0;
+#endif
 }
 
 #if ENABLE_FEATURE_FIND_PATH
 ACTF(path)
 {
+# if FNM_CASEFOLD
+	return fnmatch(ap->pattern, fileName, (ap->ipath ? FNM_CASEFOLD : 0)) == 0;
+# else
+	if (ap->ipath)
+		fileName = strcpy_upcase(alloca(strlen(fileName) + 1), fileName);
 	return fnmatch(ap->pattern, fileName, 0) == 0;
+# endif
 }
 #endif
 #if ENABLE_FEATURE_FIND_REGEX
@@ -586,16 +724,15 @@ ACTF(links)
 
 static int FAST_FUNC fileAction(const char *fileName,
 		struct stat *statbuf,
-		void *userData IF_NOT_FEATURE_FIND_MAXDEPTH(UNUSED_PARAM),
+		void *userData UNUSED_PARAM,
 		int depth IF_NOT_FEATURE_FIND_MAXDEPTH(UNUSED_PARAM))
 {
 	int r;
-#if ENABLE_FEATURE_FIND_MAXDEPTH
-#define minmaxdepth ((int*)userData)
 
-	if (depth < minmaxdepth[0])
+#if ENABLE_FEATURE_FIND_MAXDEPTH
+	if (depth < G.minmaxdepth[0])
 		return TRUE; /* skip this, continue recursing */
-	if (depth > minmaxdepth[1])
+	if (depth > G.minmaxdepth[1])
 		return SKIP; /* stop recursing */
 #endif
 
@@ -606,7 +743,7 @@ static int FAST_FUNC fileAction(const char *fileName,
 
 #if ENABLE_FEATURE_FIND_MAXDEPTH
 	if (S_ISDIR(statbuf->st_mode)) {
-		if (depth == minmaxdepth[1])
+		if (depth == G.minmaxdepth[1])
 			return SKIP;
 	}
 #endif
@@ -629,7 +766,6 @@ static int FAST_FUNC fileAction(const char *fileName,
 	/* Cannot return 0: our caller, recursive_action(),
 	 * will perror() and skip dirs (if called on dir) */
 	return (r & SKIP) ? SKIP : TRUE;
-#undef minmaxdepth
 }
 
 
@@ -674,6 +810,9 @@ static const char* plus_minus_num(const char* str)
 static action*** parse_params(char **argv)
 {
 	enum {
+	                        OPT_FOLLOW     ,
+	IF_FEATURE_FIND_XDEV(   OPT_XDEV       ,)
+	IF_FEATURE_FIND_DEPTH(  OPT_DEPTH      ,)
 	                        PARM_a         ,
 	                        PARM_o         ,
 	IF_FEATURE_FIND_NOT(	PARM_char_not  ,)
@@ -684,15 +823,20 @@ static action*** parse_params(char **argv)
 #endif
 	                        PARM_print     ,
 	IF_FEATURE_FIND_PRINT0( PARM_print0    ,)
-	IF_FEATURE_FIND_DEPTH(  PARM_depth     ,)
 	IF_FEATURE_FIND_PRUNE(  PARM_prune     ,)
 	IF_FEATURE_FIND_DELETE( PARM_delete    ,)
 	IF_FEATURE_FIND_EXEC(   PARM_exec      ,)
 	IF_FEATURE_FIND_PAREN(  PARM_char_brace,)
-	/* All options starting from here require argument */
+	/* All options/actions starting from here require argument */
 	                        PARM_name      ,
 	                        PARM_iname     ,
 	IF_FEATURE_FIND_PATH(   PARM_path      ,)
+#if ENABLE_DESKTOP
+	/* -wholename is a synonym for -path */
+	/* We support it because Linux kernel's "make tags" uses it */
+	IF_FEATURE_FIND_PATH(   PARM_wholename ,)
+#endif
+	IF_FEATURE_FIND_PATH(   PARM_ipath     ,)
 	IF_FEATURE_FIND_REGEX(  PARM_regex     ,)
 	IF_FEATURE_FIND_TYPE(   PARM_type      ,)
 	IF_FEATURE_FIND_PERM(   PARM_perm      ,)
@@ -705,28 +849,35 @@ static action*** parse_params(char **argv)
 	IF_FEATURE_FIND_SIZE(   PARM_size      ,)
 	IF_FEATURE_FIND_CONTEXT(PARM_context   ,)
 	IF_FEATURE_FIND_LINKS(  PARM_links     ,)
+	IF_FEATURE_FIND_MAXDEPTH(OPT_MINDEPTH,OPT_MAXDEPTH,)
 	};
 
 	static const char params[] ALIGN1 =
-	                         "-a\0"
-	                         "-o\0"
+	                        "-follow\0"
+	IF_FEATURE_FIND_XDEV(   "-xdev\0"                 )
+	IF_FEATURE_FIND_DEPTH(  "-depth\0"                )
+	                        "-a\0"
+	                        "-o\0"
 	IF_FEATURE_FIND_NOT(    "!\0"       )
 #if ENABLE_DESKTOP
-	                         "-and\0"
-	                         "-or\0"
-	IF_FEATURE_FIND_NOT(	 "-not\0"    )
+	                        "-and\0"
+	                        "-or\0"
+	IF_FEATURE_FIND_NOT(	"-not\0"    )
 #endif
-	                         "-print\0"
+	                        "-print\0"
 	IF_FEATURE_FIND_PRINT0( "-print0\0" )
-	IF_FEATURE_FIND_DEPTH(  "-depth\0"  )
 	IF_FEATURE_FIND_PRUNE(  "-prune\0"  )
 	IF_FEATURE_FIND_DELETE( "-delete\0" )
 	IF_FEATURE_FIND_EXEC(   "-exec\0"   )
 	IF_FEATURE_FIND_PAREN(  "(\0"       )
-	/* All options starting from here require argument */
+	/* All options/actions starting from here require argument */
 	                         "-name\0"
 	                         "-iname\0"
 	IF_FEATURE_FIND_PATH(   "-path\0"   )
+#if ENABLE_DESKTOP
+	IF_FEATURE_FIND_PATH(   "-wholename\0")
+#endif
+	IF_FEATURE_FIND_PATH(   "-ipath\0"  )
 	IF_FEATURE_FIND_REGEX(  "-regex\0"  )
 	IF_FEATURE_FIND_TYPE(   "-type\0"   )
 	IF_FEATURE_FIND_PERM(   "-perm\0"   )
@@ -739,7 +890,8 @@ static action*** parse_params(char **argv)
 	IF_FEATURE_FIND_SIZE(   "-size\0"   )
 	IF_FEATURE_FIND_CONTEXT("-context\0")
 	IF_FEATURE_FIND_LINKS(  "-links\0"  )
-	                         ;
+	IF_FEATURE_FIND_MAXDEPTH("-mindepth\0""-maxdepth\0")
+	;
 
 	action*** appp;
 	unsigned cur_group = 0;
@@ -766,27 +918,15 @@ static action*** parse_params(char **argv)
 
 	appp = xzalloc(2 * sizeof(appp[0])); /* appp[0],[1] == NULL */
 
-/* Actions have side effects and return a true or false value
- * We implement: -print, -print0, -exec
- *
- * The rest are tests.
- *
- * Tests and actions are grouped by operators
- * ( expr )              Force precedence
- * ! expr                True if expr is false
- * -not expr             Same as ! expr
- * expr1 [-a[nd]] expr2  And; expr2 is not evaluated if expr1 is false
- * expr1 -o[r] expr2     Or; expr2 is not evaluated if expr1 is true
- * expr1 , expr2         List; both expr1 and expr2 are always evaluated
- * We implement: (), -a, -o
- */
 	while (*argv) {
 		const char *arg = argv[0];
 		int parm = index_in_strings(params, arg);
 		const char *arg1 = argv[1];
 
+		dbg("arg:'%s' arg1:'%s' parm:%d PARM_type:%d", arg, arg1, parm, PARM_type);
+
 		if (parm >= PARM_name) {
-			/* All options starting from -name require argument */
+			/* All options/actions starting from -name require argument */
 			if (!arg1)
 				bb_error_msg_and_die(bb_msg_requires_arg, arg);
 			argv++;
@@ -795,11 +935,49 @@ static action*** parse_params(char **argv)
 		/* We can use big switch() here, but on i386
 		 * it doesn't give smaller code. Other arches? */
 
-	/* --- Operators --- */
-		if (parm == PARM_a IF_DESKTOP(|| parm == PARM_and)) {
+/* Options always return true. They always take effect
+ * rather than being processed only when their place in the
+ * expression is reached.
+ */
+		/* Options */
+		if (parm == OPT_FOLLOW) {
+                        dbg("follow enabled: %d", __LINE__);
+                        G.recurse_flags |= ACTION_FOLLOWLINKS | ACTION_DANGLING_OK;
+		}
+#if ENABLE_FEATURE_FIND_XDEV
+		else if (parm == OPT_XDEV) {
+			dbg("%d", __LINE__);
+			G.xdev_on = 1;
+		}
+#endif
+#if ENABLE_FEATURE_FIND_MAXDEPTH
+		else if (parm == OPT_MINDEPTH || parm == OPT_MINDEPTH + 1) {
+			dbg("%d", __LINE__);
+			G.minmaxdepth[parm - OPT_MINDEPTH] = xatoi_positive(arg1);
+		}
+#endif
+#if ENABLE_FEATURE_FIND_DEPTH
+		else if (parm == OPT_DEPTH) {
+			dbg("%d", __LINE__);
+			G.recurse_flags |= ACTION_DEPTHFIRST;
+		}
+#endif
+/* Actions are grouped by operators
+ * ( expr )              Force precedence
+ * ! expr                True if expr is false
+ * -not expr             Same as ! expr
+ * expr1 [-a[nd]] expr2  And; expr2 is not evaluated if expr1 is false
+ * expr1 -o[r] expr2     Or; expr2 is not evaluated if expr1 is true
+ * expr1 , expr2         List; both expr1 and expr2 are always evaluated
+ * We implement: (), -a, -o
+ */
+		/* Operators */
+		else if (parm == PARM_a IF_DESKTOP(|| parm == PARM_and)) {
+			dbg("%d", __LINE__);
 			/* no further special handling required */
 		}
 		else if (parm == PARM_o IF_DESKTOP(|| parm == PARM_or)) {
+			dbg("%d", __LINE__);
 			/* start new OR group */
 			cur_group++;
 			appp = xrealloc(appp, (cur_group+2) * sizeof(*appp));
@@ -811,36 +989,31 @@ static action*** parse_params(char **argv)
 		else if (parm == PARM_char_not IF_DESKTOP(|| parm == PARM_not)) {
 			/* also handles "find ! ! -name 'foo*'" */
 			invert_flag ^= 1;
+			dbg("invert_flag:%d", invert_flag);
 		}
 #endif
-
-	/* --- Tests and actions --- */
+		/* Actions */
 		else if (parm == PARM_print) {
+			dbg("%d", __LINE__);
 			G.need_print = 0;
-			/* GNU find ignores '!' here: "find ! -print" */
-			IF_FEATURE_FIND_NOT( invert_flag = 0; )
 			(void) ALLOC_ACTION(print);
 		}
 #if ENABLE_FEATURE_FIND_PRINT0
 		else if (parm == PARM_print0) {
+			dbg("%d", __LINE__);
 			G.need_print = 0;
-			IF_FEATURE_FIND_NOT( invert_flag = 0; )
 			(void) ALLOC_ACTION(print0);
 		}
 #endif
-#if ENABLE_FEATURE_FIND_DEPTH
-		else if (parm == PARM_depth) {
-			G.recurse_flags |= ACTION_DEPTHFIRST;
-		}
-#endif
 #if ENABLE_FEATURE_FIND_PRUNE
 		else if (parm == PARM_prune) {
-			IF_FEATURE_FIND_NOT( invert_flag = 0; )
+			dbg("%d", __LINE__);
 			(void) ALLOC_ACTION(prune);
 		}
 #endif
 #if ENABLE_FEATURE_FIND_DELETE
 		else if (parm == PARM_delete) {
+			dbg("%d", __LINE__);
 			G.need_print = 0;
 			G.recurse_flags |= ACTION_DEPTHFIRST;
 			(void) ALLOC_ACTION(delete);
@@ -850,8 +1023,8 @@ static action*** parse_params(char **argv)
 		else if (parm == PARM_exec) {
 			int i;
 			action_exec *ap;
+			dbg("%d", __LINE__);
 			G.need_print = 0;
-			IF_FEATURE_FIND_NOT( invert_flag = 0; )
 			ap = ALLOC_ACTION(exec);
 			ap->exec_argv = ++argv; /* first arg after -exec */
 			/*ap->exec_argc = 0; - ALLOC_ACTION did it */
@@ -859,9 +1032,9 @@ static action*** parse_params(char **argv)
 				if (!*argv) /* did not see ';' or '+' until end */
 					bb_error_msg_and_die(bb_msg_requires_arg, "-exec");
 				// find -exec echo Foo ">{}<" ";"
-				// executes "echo Foo <filename>",
+				// executes "echo Foo >FILENAME<",
 				// find -exec echo Foo ">{}<" "+"
-				// executes "echo Foo <filename1> <filename2> <filename3>...".
+				// executes "echo Foo FILENAME1 FILENAME2 FILENAME3...".
 				// TODO (so far we treat "+" just like ";")
 				if ((argv[0][0] == ';' || argv[0][0] == '+')
 				 && argv[0][1] == '\0'
@@ -885,6 +1058,7 @@ static action*** parse_params(char **argv)
 			char **endarg;
 			unsigned nested = 1;
 
+			dbg("%d", __LINE__);
 			endarg = argv;
 			while (1) {
 				if (!*++endarg)
@@ -904,20 +1078,24 @@ static action*** parse_params(char **argv)
 #endif
 		else if (parm == PARM_name || parm == PARM_iname) {
 			action_name *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(name);
 			ap->pattern = arg1;
 			ap->iname = (parm == PARM_iname);
 		}
 #if ENABLE_FEATURE_FIND_PATH
-		else if (parm == PARM_path) {
+		else if (parm == PARM_path IF_DESKTOP(|| parm == PARM_wholename) || parm == PARM_ipath) {
 			action_path *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(path);
 			ap->pattern = arg1;
+			ap->ipath = (parm == PARM_ipath);
 		}
 #endif
 #if ENABLE_FEATURE_FIND_REGEX
 		else if (parm == PARM_regex) {
 			action_regex *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(regex);
 			xregcomp(&ap->compiled_pattern, arg1, 0 /*cflags*/);
 		}
@@ -927,16 +1105,18 @@ static action*** parse_params(char **argv)
 			action_type *ap;
 			ap = ALLOC_ACTION(type);
 			ap->type_mask = find_type(arg1);
+			dbg("created:type mask:%x", ap->type_mask);
 		}
 #endif
 #if ENABLE_FEATURE_FIND_PERM
-/* -perm mode   File's permission bits are exactly mode (octal or symbolic).
+/* -perm BITS   File's mode bits are exactly BITS (octal or symbolic).
  *              Symbolic modes use mode 0 as a point of departure.
- * -perm -mode  All of the permission bits mode are set for the file.
- * -perm +mode  Any of the permission bits mode are set for the file.
+ * -perm -BITS  All of the BITS are set in file's mode.
+ * -perm +BITS  At least one of the BITS is set in file's mode.
  */
 		else if (parm == PARM_perm) {
 			action_perm *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(perm);
 			ap->perm_char = arg1[0];
 			arg1 = plus_minus_num(arg1);
@@ -948,6 +1128,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_MTIME
 		else if (parm == PARM_mtime) {
 			action_mtime *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(mtime);
 			ap->mtime_char = arg1[0];
 			ap->mtime_days = xatoul(plus_minus_num(arg1));
@@ -956,6 +1137,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_MMIN
 		else if (parm == PARM_mmin) {
 			action_mmin *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(mmin);
 			ap->mmin_char = arg1[0];
 			ap->mmin_mins = xatoul(plus_minus_num(arg1));
@@ -965,6 +1147,7 @@ static action*** parse_params(char **argv)
 		else if (parm == PARM_newer) {
 			struct stat stat_newer;
 			action_newer *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(newer);
 			xstat(arg1, &stat_newer);
 			ap->newer_mtime = stat_newer.st_mtime;
@@ -973,6 +1156,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_INUM
 		else if (parm == PARM_inum) {
 			action_inum *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(inum);
 			ap->inode_num = xatoul(arg1);
 		}
@@ -980,6 +1164,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_USER
 		else if (parm == PARM_user) {
 			action_user *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(user);
 			ap->uid = bb_strtou(arg1, NULL, 10);
 			if (errno)
@@ -989,6 +1174,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_GROUP
 		else if (parm == PARM_group) {
 			action_group *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(group);
 			ap->gid = bb_strtou(arg1, NULL, 10);
 			if (errno)
@@ -1017,6 +1203,7 @@ static action*** parse_params(char **argv)
 				{ "", 0 }
 			};
 			action_size *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(size);
 			ap->size_char = arg1[0];
 			ap->size = XATOU_SFX(plus_minus_num(arg1), find_suffixes);
@@ -1025,6 +1212,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_CONTEXT
 		else if (parm == PARM_context) {
 			action_context *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(context);
 			/*ap->context = NULL; - ALLOC_ACTION did it */
 			/* SELinux headers erroneously declare non-const parameter */
@@ -1035,6 +1223,7 @@ static action*** parse_params(char **argv)
 #if ENABLE_FEATURE_FIND_LINKS
 		else if (parm == PARM_links) {
 			action_links *ap;
+			dbg("%d", __LINE__);
 			ap = ALLOC_ACTION(links);
 			ap->links_char = arg1[0];
 			ap->links_count = xatoul(plus_minus_num(arg1));
@@ -1046,197 +1235,64 @@ static action*** parse_params(char **argv)
 		}
 		argv++;
 	}
+	dbg("exiting %s", __func__);
 	return appp;
 #undef ALLOC_ACTION
 }
 
-//usage:#define find_trivial_usage
-//usage:       "[PATH]... [EXPRESSION]"
-//usage:#define find_full_usage "\n\n"
-//usage:       "Search for files. The default PATH is the current directory,\n"
-//usage:       "default EXPRESSION is '-print'\n"
-//usage:     "\nEXPRESSION may consist of:"
-//usage:     "\n	-follow		Follow symlinks"
-//usage:	IF_FEATURE_FIND_XDEV(
-//usage:     "\n	-xdev		Don't descend directories on other filesystems"
-//usage:	)
-//usage:	IF_FEATURE_FIND_MAXDEPTH(
-//usage:     "\n	-maxdepth N	Descend at most N levels. -maxdepth 0 applies"
-//usage:     "\n			tests/actions to command line arguments only"
-//usage:	)
-//usage:     "\n	-mindepth N	Don't act on first N levels"
-//usage:     "\n	-name PATTERN	File name (w/o directory name) matches PATTERN"
-//usage:     "\n	-iname PATTERN	Case insensitive -name"
-//usage:	IF_FEATURE_FIND_PATH(
-//usage:     "\n	-path PATTERN	Path matches PATTERN"
-//usage:	)
-//usage:	IF_FEATURE_FIND_REGEX(
-//usage:     "\n	-regex PATTERN	Path matches regex PATTERN"
-//usage:	)
-//usage:	IF_FEATURE_FIND_TYPE(
-//usage:     "\n	-type X		File type is X (X is one of: f,d,l,b,c,...)"
-//usage:	)
-//usage:	IF_FEATURE_FIND_PERM(
-//usage:     "\n	-perm NNN	Permissions match any of (+NNN), all of (-NNN),"
-//usage:     "\n			or exactly NNN"
-//usage:	)
-//usage:	IF_FEATURE_FIND_MTIME(
-//usage:     "\n	-mtime DAYS	Modified time is greater than (+N), less than (-N),"
-//usage:     "\n			or exactly N days"
-//usage:	)
-//usage:	IF_FEATURE_FIND_MMIN(
-//usage:     "\n	-mmin MINS	Modified time is greater than (+N), less than (-N),"
-//usage:     "\n			or exactly N minutes"
-//usage:	)
-//usage:	IF_FEATURE_FIND_NEWER(
-//usage:     "\n	-newer FILE	Modified time is more recent than FILE's"
-//usage:	)
-//usage:	IF_FEATURE_FIND_INUM(
-//usage:     "\n	-inum N		File has inode number N"
-//usage:	)
-//usage:	IF_FEATURE_FIND_USER(
-//usage:     "\n	-user NAME	File is owned by user NAME (numeric user ID allowed)"
-//usage:	)
-//usage:	IF_FEATURE_FIND_GROUP(
-//usage:     "\n	-group NAME	File belongs to group NAME (numeric group ID allowed)"
-//usage:	)
-//usage:	IF_FEATURE_FIND_DEPTH(
-//usage:     "\n	-depth		Process directory name after traversing it"
-//usage:	)
-//usage:	IF_FEATURE_FIND_SIZE(
-//usage:     "\n	-size N[bck]	File size is N (c:bytes,k:kbytes,b:512 bytes(def.))"
-//usage:     "\n			+/-N: file size is bigger/smaller than N"
-//usage:	)
-//usage:	IF_FEATURE_FIND_LINKS(
-//usage:     "\n	-links N	Number of links is greater than (+N), less than (-N),"
-//usage:     "\n			or exactly N"
-//usage:	)
-//usage:     "\n	-print		Print (default and assumed)"
-//usage:	IF_FEATURE_FIND_PRINT0(
-//usage:     "\n	-print0		Delimit output with null characters rather than"
-//usage:     "\n			newlines"
-//usage:	)
-//usage:	IF_FEATURE_FIND_CONTEXT(
-//usage:     "\n	-context	File has specified security context"
-//usage:	)
-//usage:	IF_FEATURE_FIND_EXEC(
-//usage:     "\n	-exec CMD ARG ;	Run CMD with all instances of {} replaced by the"
-//usage:     "\n			matching files"
-//usage:	)
-//usage:	IF_FEATURE_FIND_PRUNE(
-//usage:     "\n	-prune		Stop traversing current subtree"
-//usage:	)
-//usage:	IF_FEATURE_FIND_DELETE(
-//usage:     "\n	-delete		Delete files, turns on -depth option"
-//usage:	)
-//usage:	IF_FEATURE_FIND_PAREN(
-//usage:     "\n	(EXPR)		Group an expression"
-//usage:	)
-//usage:
-//usage:#define find_example_usage
-//usage:       "$ find / -name passwd\n"
-//usage:       "/etc/passwd\n"
-
 int find_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
 int find_main(int argc UNUSED_PARAM, char **argv)
 {
-	static const char options[] ALIGN1 =
-	                  "-follow\0"
-IF_FEATURE_FIND_XDEV(    "-xdev\0"    )
-IF_FEATURE_FIND_MAXDEPTH("-mindepth\0""-maxdepth\0")
-	                  ;
-	enum {
-	                  OPT_FOLLOW,
-IF_FEATURE_FIND_XDEV(    OPT_XDEV    ,)
-IF_FEATURE_FIND_MAXDEPTH(OPT_MINDEPTH,)
-	};
-
-	char *arg;
-	char **argp;
 	int i, firstopt, status = EXIT_SUCCESS;
-#if ENABLE_FEATURE_FIND_MAXDEPTH
-	int minmaxdepth[2] = { 0, INT_MAX };
-#else
-#define minmaxdepth NULL
-#endif
 
 	INIT_G();
 
-	for (firstopt = 1; argv[firstopt]; firstopt++) {
+	argv++;
+	for (firstopt = 0; argv[firstopt]; firstopt++) {
 		if (argv[firstopt][0] == '-')
 			break;
 		if (ENABLE_FEATURE_FIND_NOT && LONE_CHAR(argv[firstopt], '!'))
 			break;
-#if ENABLE_FEATURE_FIND_PAREN
-		if (LONE_CHAR(argv[firstopt], '('))
+		if (ENABLE_FEATURE_FIND_PAREN && LONE_CHAR(argv[firstopt], '('))
 			break;
-#endif
 	}
-	if (firstopt == 1) {
-		argv[0] = (char*)".";
-		argv--;
+	if (firstopt == 0) {
+		*--argv = (char*)".";
 		firstopt++;
 	}
 
-/* All options always return true. They always take effect
- * rather than being processed only when their place in the
- * expression is reached.
- * We implement: -follow, -xdev, -maxdepth
- */
-	/* Process options, and replace then with -a */
-	/* (-a will be ignored by recursive parser later) */
-	argp = &argv[firstopt];
-	while ((arg = argp[0])) {
-		int opt = index_in_strings(options, arg);
-		if (opt == OPT_FOLLOW) {
-			G.recurse_flags |= ACTION_FOLLOWLINKS | ACTION_DANGLING_OK;
-			argp[0] = (char*)"-a";
-		}
-#if ENABLE_FEATURE_FIND_XDEV
-		if (opt == OPT_XDEV) {
-			struct stat stbuf;
-			if (!G.xdev_count) {
-				G.xdev_count = firstopt - 1;
-				G.xdev_dev = xzalloc(G.xdev_count * sizeof(G.xdev_dev[0]));
-				for (i = 1; i < firstopt; i++) {
-					/* not xstat(): shouldn't bomb out on
-					 * "find not_exist exist -xdev" */
-					if (stat(argv[i], &stbuf) == 0)
-						G.xdev_dev[i-1] = stbuf.st_dev;
-					/* else G.xdev_dev[i-1] stays 0 and
-					 * won't match any real device dev_t */
-				}
-			}
-			argp[0] = (char*)"-a";
-		}
-#endif
-#if ENABLE_FEATURE_FIND_MAXDEPTH
-		if (opt == OPT_MINDEPTH || opt == OPT_MINDEPTH + 1) {
-			if (!argp[1])
-				bb_show_usage();
-			minmaxdepth[opt - OPT_MINDEPTH] = xatoi_positive(argp[1]);
-			argp[0] = (char*)"-a";
-			argp[1] = (char*)"-a";
-			argp++;
-		}
-#endif
-		argp++;
-	}
-
 	G.actions = parse_params(&argv[firstopt]);
+	argv[firstopt] = NULL;
 
-	for (i = 1; i < firstopt; i++) {
+#if ENABLE_FEATURE_FIND_XDEV
+	if (G.xdev_on) {
+		struct stat stbuf;
+
+		G.xdev_count = firstopt;
+		G.xdev_dev = xzalloc(G.xdev_count * sizeof(G.xdev_dev[0]));
+		for (i = 0; argv[i]; i++) {
+			/* not xstat(): shouldn't bomb out on
+			 * "find not_exist exist -xdev" */
+			if (stat(argv[i], &stbuf) == 0)
+				G.xdev_dev[i] = stbuf.st_dev;
+			/* else G.xdev_dev[i] stays 0 and
+			 * won't match any real device dev_t
+			 */
+		}
+	}
+#endif
+
+	for (i = 0; argv[i]; i++) {
 		if (!recursive_action(argv[i],
 				G.recurse_flags,/* flags */
 				fileAction,     /* file action */
 				fileAction,     /* dir action */
-#if ENABLE_FEATURE_FIND_MAXDEPTH
-				minmaxdepth,    /* user data */
-#else
 				NULL,           /* user data */
-#endif
-				0))             /* depth */
+				0)              /* depth */
+		) {
 			status = EXIT_FAILURE;
+		}
 	}
+
 	return status;
 }
