@@ -9,14 +9,23 @@
  *
  */
 
+//usage:#define rdev_trivial_usage
+//usage:       ""
+//usage:#define rdev_full_usage "\n\n"
+//usage:       "Print the device node associated with the filesystem mounted at '/'"
+//usage:
+//usage:#define rdev_example_usage
+//usage:       "$ rdev\n"
+//usage:       "/dev/mtdblock9 /\n"
+
 #include "libbb.h"
 
 int rdev_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
 int rdev_main(int argc UNUSED_PARAM, char **argv UNUSED_PARAM)
 {
-	char const * const root_device = find_block_device("/");
+	const char *root_device = find_block_device("/");
 
-	if (root_device != NULL) {
+	if (root_device) {
 		printf("%s /\n", root_device);
 		return EXIT_SUCCESS;
 	}
